/***************************************************************************
 * athcool (Athlon Powersaving utility)

athcool is a small utility, enabling/disabling Powersaving mode
for AMD Athlon/Duron processors.

Since enabling Powersaving mode, you can save power consumption, 
lower CPU temprature when CPU is idle.

Powersaving works if your kernel support ACPI (APM not work),
because athcool only set/unset "Disconnect enable when STPGNT detected"
bits in the Northbridge of Chipset.
To really save power, someone has to send the STPGNT signal when idle.
This is done by the ACPI subsystem when C2 state enterd.

!!!WARNING!!!
Depending on your motherboard and/or hardware components, 
enabling Athlon powersaving mode may cause:

 * noisy or distorted sound playback
 * a slowdown in harddisk performance
 * system locks or instability

If you met those problems, you should not use athcool.
Please use athcool AT YOUR OWN RISK.

   Osamu Kayasono <jacobi@jcom.home.ne.jp>
   http://members.jcom.home.ne.jp/jacobi/linux/softwares.html

  (Related links)
 * CoolOn Project [ written in Japanese ]
	http://homepage2.nifty.com/coolon/
 * pciutils
	http://atrey.karlin.mff.cuni.cz/~mj/pciutils.shtml
 * Athlon Powersaving HOWTO
	http://cip.uni-trier.de/nofftz/linux/Athlon-Powersaving-HOWTO.html
 * LVCool
	http://vcool.occludo.net/

 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#include <stdlib.h>
#include <unistd.h>
#include <stdio.h>
#include <stdint.h>
#include <sys/io.h>
#include <string.h>
#include <pci/pci.h>
#include  "athcool.h"
#include  "chips.h"

extern struct pci_filter filter;
extern struct pci_access *pacc;
extern struct device *scan_devices( void );

static void print_title( void )
{
  printf( PACKAGE " version " VERSION );
  printf( " - enabling/disabling Athlon Powersaving mode\n\n");
}

static void print_warnings( void )
{
  printf( "!!!WARNING!!!\n");
  printf( "Depending on your motherboard and/or hardware components, \n" );
  printf( "enabling Athlon powersaving mode may cause:\n" );
  printf( " * noisy or distorted sound playback\n" );
  printf( " * a slowdown in harddisk performance\n" );
  printf( " * system locks or instability\n\n" );
  printf( "Please use athcool AT YOUR OWN RISK.\n\n" );
}

static void list_chipsets( void )
{
  AthlonChips_s	*p_athlon;

  printf( "  Supported Chipsets:\n" );
  printf( "\t[Reg ID]  chipsets name\n" );
  for ( p_athlon = AthlonChips; p_athlon->name != NULL; p_athlon++ ) {
    printf( "\t[ 0x%02x ]  %s\n", p_athlon->chipreg, p_athlon->name );
  }
}

static void usage( const char *package )
{
  printf( " Usage: %s option\n", package );
  printf( "  supported options\n" );
  printf( "\ton\t enable Athlon Powersaving mode\n" );
  printf( "\toff\t disable Athlon Powersaving mode\n" );
  printf( "\tstat\t query Athlon Powersaving mode\n" );
  printf( "\tlist\t list supported chipsets\n\n" );

#ifdef BUILD_EXPERIMENTAL
  printf( " %s on|off|stat force RegID\n", package );
  printf( "\tmanual specification of chipsets\n" );
  printf( "\tDANGEROUS. Must not use this option, unless it is sure\n" );
  printf( "\tthat it has same config as your chipsets.\n\n" );
#endif
}

int setAthlonPowersave( int m, int force_id, struct device *first_dev )
{
  unsigned int	regval_cur, regval_new, res;
  struct device	*d = NULL;
  const char	*msg;
  AthlonChips_s	*p_athlon = NULL;
  ChipReg_s	*p_reg;
  EachReg_s	*p_eachreg;

#ifdef BUILD_EXPERIMENTAL
  if (force_id >= 0) {
    d = first_dev;
    /* search forced Reg ID */
    for ( p_reg = ChipReg; p_reg->regs->desc != NULL; p_reg++ ) {
      if ( p_reg->regid == force_id ) { break; }
    }
    m ^=  ATHLONPM_FORCE;
    printf( " *** disabled auto detection on user request.\n" );
  } else {
#endif
    /* search supported chipset */
    for ( d = first_dev; d != NULL; d = d->next ) {
      for ( p_athlon = AthlonChips; p_athlon->name != NULL ; p_athlon++ ) {
	if ( (d->dev->vendor_id == p_athlon->vendorid) &&
	     (d->dev->device_id == p_athlon->deviceid) ) {
	  printf( "%s (%04x %04x) found\n", p_athlon->name,
		  d->dev->vendor_id, d->dev->device_id );
	  break;
	}
      }
      if ( p_athlon->name != NULL ) { break; }
    }
    
    if ( p_athlon == NULL || p_athlon->name == NULL ) {
      printf( "no supported Chipset found. nothing changed.\n" );
      return 1;
    }

    /* search chipset config */
    for ( p_reg = ChipReg; p_reg->regs->desc != NULL; p_reg++ ) {
      if ( p_reg->regid == p_athlon->chipreg ) { break; }
    }
#ifdef BUILD_EXPERIMENTAL
  }
#endif
  if ( p_reg->regs->desc == NULL ) {
    printf( "internal error."
	    " supported Chipset found,"
	    " but Chipset Configuration not found.\n" );
    return 1;
  }

  /* read current value of register */
  for ( p_eachreg = p_reg->regs; p_eachreg->desc != NULL; p_eachreg++ ) {
    switch (p_reg->tp) {
    case BYTE:
      regval_cur = pci_read_byte( d->dev, p_eachreg->addr );
      break;
    case SHORT:
      regval_cur = pci_read_word( d->dev, p_eachreg->addr );
      break;
    case LONG:
    default:
      regval_cur = pci_read_long( d->dev, p_eachreg->addr );
      break;
    }

    switch (m) {
    case ATHLONPM_STAT:
      /* begin of checking status */
      printf( "'%s' bit is %s.\n", p_eachreg->desc,
	      ( (regval_cur & p_eachreg->bit) == p_eachreg->bit ) ?
	      "enabled" : "disabled" );
      break;
      /* end of checking status */
    case ATHLONPM_ON:
    case ATHLONPM_OFF:
      /* begin of enabling/disabling bits */
      regval_new = regval_cur | p_eachreg->bit;
      msg = ( m == ATHLONPM_ON ) ? "enabl" : "disabl";
      printf( "%sing '%s' bit ... ", msg, p_eachreg->desc );
      
      if ( ((regval_cur == regval_new) && (m == ATHLONPM_ON)) ||
	   ((regval_cur != regval_new) && (m == ATHLONPM_OFF)) ) {
	printf( "already %sed.\n", msg );
	break;
      }

      if ( m == ATHLONPM_OFF ) { regval_new = regval_cur ^ p_eachreg->bit; }

      /* write new value into register */
      switch (p_reg->tp) {
      case BYTE:
#ifndef ATHLON_PM_READONLY
	res = pci_write_byte( d->dev, p_eachreg->addr, (uint8_t) regval_new );
#endif
	printf( " done\n" );
	printf( " Address 0x%02X : 0x%02X -> 0x%02X\n", 
	      p_eachreg->addr, regval_cur, regval_new );
	break;
      case SHORT:
#ifndef ATHLON_PM_READONLY
	res = pci_write_word( d->dev, p_eachreg->addr, (uint16_t) regval_new );
#endif
	printf( " done\n" );
	/* x86 cpus are little-endian, so values are inverted in registers */
	printf( " Address 0x%02X - 0x%02X : 0x%04X -> 0x%04X\n", 
		p_eachreg->addr+1, p_eachreg->addr, regval_cur, regval_new );
	break;
      case LONG:
      default:
#ifndef ATHLON_PM_READONLY
	res = pci_write_long( d->dev, p_eachreg->addr, (uint32_t) regval_new );
#endif
	printf( " done\n" );
	/* x86 cpus are little-endian, so values are inverted in registers */
	printf( " Address 0x%02X - 0x%02X : 0x%08X -> 0x%08X\n", 
		p_eachreg->addr+3, p_eachreg->addr, regval_cur, regval_new );
	break;
      }
    
      break;
      /* end of enabling/disabling bits */
    }
  }

  return 0;
}

#ifdef BUILD_EXPERIMENTAL
int hex2int(char c)
{ 
  int result;

  if     ( '0' <= c && c <= '9'){ result = c - '0';      } // 0〜9
  else if( 'a' <= c && c <= 'f'){ result = c - 'a' + 10; } // a〜f
  else if( 'A' <= c && c <= 'F'){ result = c - 'A' + 10; } // A〜F
  else                          { result = -1; }           // Others

  return result;
}

int getforceid ( char *idstr )
{
  int result = -1;

  if (strlen(idstr)<2) {
    result = hex2int(idstr[0]);
  } else {
    result = hex2int(idstr[0]) * 16 + hex2int(idstr[1]);
  }

  return result;
}
#endif

int main ( int argc, char *argv[] )
{
  int	mode = ATHLONPM_USAGE;
  int	force_id = -1;
  int	result, c;

  /* read optarg */
  for (c = 1; c < argc; c++ ) {
    if ( strcmp( argv[c], "on") == 0 ) {
      mode = ATHLONPM_ON;
    } else if ( strcmp( argv[c], "off" ) == 0 ) {
      mode = ATHLONPM_OFF;
    } else if ( strcmp( argv[c], "stat" ) == 0 ) {
      mode = ATHLONPM_STAT;
    } else if ( strcmp( argv[c], "list" ) == 0 ) {
      mode = ATHLONPM_LIST;
#ifdef BUILD_EXPERIMENTAL
    } else if ( strcmp( argv[c], "force" ) == 0 ) {
      c++;
      if (c < argc) force_id = getforceid( argv[c] );
#endif
    }
  }

  print_title();

  switch (mode) {
  case ATHLONPM_USAGE:
    usage( argv[0] );
    exit (1);
  case ATHLONPM_LIST:
    list_chipsets();
    exit (1);
  }

#ifdef BUILD_EXPERIMENTAL
  if ( force_id >= 0) { mode |= ATHLONPM_FORCE; }
#endif
  print_warnings();

  if ( iopl( 3 ) < 0 ) {
    printf( "%s : must run as root. exit\n", argv[0] );
    exit (1);
  }

  /* init pciutil, and scan devices */
  pacc = pci_alloc();
  pci_filter_init( pacc, &filter );
  pci_init( pacc );

  result = setAthlonPowersave( mode, force_id, scan_devices() );

  /* clean up, exit */
  pci_cleanup( pacc );

  return result;
}
/* End Of File	**************************************************************/
